(in-package "USER")

(unless (find-package "DSYS")
  (make-package "DSYS" :use '("LISP")))

;(push :testing *features*)

#|| #+cmu
(progn

(setf (search-list "lisp:") '("/usr16/dice/rharris/cmucl/build/"
			      "/usr20/dice/rharris/lisp/"
			      "/usr20/dice/rharris/cmucl/source/"))

(let ((sym (intern "*DEFAULT-FILE-LOCATION*" "DSYS")))
  (proclaim `(special ,sym))
  (set sym "lisp:system"))

) ||#

(unless (and (find-symbol "INITIALIZE" "DSYS")
	     (fboundp (find-symbol "INITIALIZE" "DSYS")))
  (proclaim '(special *initialize-systems-p*))
  (setq *initialize-systems-p* nil)
  (load #-cmu "system"
	#+cmu "lisp:system")
  (setq *initialize-systems-p* t))

(when (< (length dsys::*all-systems*) 2)
  (dsys:initialize-systems :system-only-p (not (y-or-n-p "Load systems? "))))


