;;; -*- Mode: Lisp; Package: DSYS; Base: 10; Syntax: Common-Lisp -*-

(in-package "DSYS")

(set-system-source-file 'inspecter
			(subfile (list "inspecter")
				 :name "sysdef"))

(pushnew 'inspecter *auto-load-systems*)
