
(in-package "LISP")

#+cmu
(eval-when (compile load eval)
  (let ((insp (find-package "INSPECT")))	
    (when (and insp
	       (member (find-package "KERNEL")
		       (package-use-list insp)))
      (rename-package insp "%INSPECT"))))		       

(in-package "INSPECT" :use '("LISP" "PCL" #-cmu "CONDITIONS"))

(export '(inspect))

(unless (find-package "ROSE")
  (make-package "ROSE" :use '("LISP")))

(unless (find-package "RS")
  (make-package "RS" ':use '()))
