
(IN-PACKAGE "USER") 
(PROCLAIM '(FTYPE (FUNCTION (*) FIXNUM) PCL::ZERO)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T FIXNUM *) FIXNUM)
            PCL::COMPUTE-PRIMARY-CACHE-LOCATION-FROM-LOCATION)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T) FIXNUM) PCL::FAST-INSTANCE-BOUNDP-INDEX
            PCL::ONE-INDEX-LIMIT-FN PCL::N-N-ACCESSORS-LIMIT-FN
            PCL::CHECKING-LIMIT-FN PCL::PV-CACHE-LIMIT-FN
            PCL::CACHE-NLINES PCL::CACHE-MAX-LOCATION PCL::CACHE-SIZE
            PCL::CACHE-MASK PCL::ARG-INFO-NUMBER-REQUIRED
            PCL::DEFAULT-LIMIT-FN PCL::CACHE-COUNT
            PCL::CACHING-LIMIT-FN PCL::PV-TABLE-PV-SIZE
            PCL::EARLY-CLASS-SIZE)) 
(PROCLAIM '(FTYPE (FUNCTION (FIXNUM) T) PCL::POWER-OF-TWO-CEILING)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T) FUNCTION) PCL::CACHE-LIMIT-FN
            PCL::METHOD-CALL-FUNCTION PCL::FAST-METHOD-CALL-FUNCTION)) 
(PROCLAIM '(FTYPE (FUNCTION (T) PCL::FIELD-TYPE) PCL::CACHE-FIELD)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T) LIST) PCL::CACHE-OVERFLOW
            PCL::PV-TABLE-SLOT-NAME-LISTS PCL::PV-TABLE-CALL-LIST)) 
(PROCLAIM '(FTYPE (FUNCTION (T) (MEMBER NIL T)) PCL::CACHE-VALUEP)) 
(PROCLAIM '(FTYPE (FUNCTION (T) SIMPLE-VECTOR) PCL::CACHE-VECTOR)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T) (VALUES T T)) PCL::MAKE-CLASS-PREDICATE-NAME
            PCL::MAKE-KEYWORD)) 
(PROCLAIM '(FTYPE (FUNCTION (FIXNUM T) T) PCL::%CCLOSURE-ENV-NTHCDR)) 
(PROCLAIM
    '(FTYPE (FUNCTION (FIXNUM FIXNUM T) FIXNUM)
            PCL::COMPUTE-PRIMARY-CACHE-LOCATION)) 
(PROCLAIM '(FTYPE (FUNCTION (T) (INTEGER 1 512)) PCL::CACHE-LINE-SIZE)) 
(PROCLAIM '(FTYPE (FUNCTION (T) (INTEGER 1 256)) PCL::CACHE-NKEYS)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T) (OR PCL::CACHE NULL)) PCL::PV-TABLE-CACHE)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T T) *) PCL::MEMF-CODE-CONVERTER
            PCL::REAL-LOAD-DEFCLASS PCL::CACHE-MISS-VALUES-INTERNAL
            PCL::GENERATE-DISCRIMINATION-NET-INTERNAL
            PCL::MAKE-LONG-METHOD-COMBINATION-FUNCTION
            PCL::DO-SHORT-METHOD-COMBINATION
            WALKER::WALK-TEMPLATE-HANDLE-REPEAT-1)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T) *)
            PCL::GET-OPTIMIZED-STD-SLOT-VALUE-USING-CLASS-METHOD-FUNCTION
            ITERATE::WALK-GATHERING-BODY
            PCL::CACHE-MISS-VALUES
            PCL::MAKE-OPTIMIZED-STD-READER-METHOD-FUNCTION
            PCL::OPTIMIZE-SLOT-VALUE-BY-CLASS-P PCL::ACCESSOR-VALUES1
            PCL::EMIT-READER/WRITER
            PCL::EMIT-ONE-OR-N-INDEX-READER/WRITER
            WALKER::WALK-MULTIPLE-VALUE-SETQ PCL::GENERATING-LISP
            PCL::EMIT-READER/WRITER-FUNCTION
            PCL::EMIT-ONE-OR-N-INDEX-READER/WRITER-FUNCTION
            WALKER::WALK-LET-IF PCL::SET-SLOT-VALUE
            PCL::CONVERT-METHODS PCL::SLOT-VALUE-USING-CLASS-DFUN
            PCL::SLOT-BOUNDP-USING-CLASS-DFUN
            PCL::CHECK-METHOD-ARG-INFO PCL::LOAD-LONG-DEFCOMBIN
            PCL::MAKE-FINAL-N-N-ACCESSOR-DFUN
            PCL::MAKE-FINAL-CACHING-DFUN
            PCL::MAKE-FINAL-CONSTANT-VALUE-DFUN
            PCL::GET-CLASS-SLOT-VALUE-1 PCL::ACCESSOR-VALUES-INTERNAL
            PCL::MAKE-OPTIMIZED-STD-WRITER-METHOD-FUNCTION
            PCL::MAKE-OPTIMIZED-STD-BOUNDP-METHOD-FUNCTION
            ITERATE::EXPAND-INTO-LET WALKER::WALK-FORM-INTERNAL
            ITERATE::RENAME-VARIABLES
            PCL::CONSTANT-VALUE-MISS PCL::CACHING-MISS
            PCL::CHECKING-MISS
            PCL::GET-OPTIMIZED-STD-ACCESSOR-METHOD-FUNCTION)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T) *)
            PCL::COMPUTE-DISCRIMINATING-FUNCTION-ARGLIST-INFO-INTERNAL
            PCL::ADD-METHOD-DECLARATIONS PCL::WALK-METHOD-LAMBDA
            PCL::MAKE-TWO-CLASS-ACCESSOR-DFUN
            WALKER::WALK-TEMPLATE-HANDLE-REPEAT)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T) *)
            PCL::GET-ACCESSOR-FROM-SVUC-METHOD-FUNCTION
            PCL::BOOTSTRAP-ACCESSOR-DEFINITION
            PCL::GET-ACCESSOR-METHOD-FUNCTION
            PCL::EMIT-CHECKING-OR-CACHING
            PCL::EMIT-CHECKING-OR-CACHING-FUNCTION
            PCL::SETF-SLOT-VALUE-USING-CLASS-DFUN
            PCL::LOAD-SHORT-DEFCOMBIN
            PCL::INITIALIZE-INSTANCE-SIMPLE-FUNCTION
            PCL::MAKE-SHARED-INITIALIZE-FORM-LIST
            PCL::MAKE-ONE-CLASS-ACCESSOR-DFUN
            PCL::MAKE-FINAL-ONE-INDEX-ACCESSOR-DFUN
            PCL::MAKE-FINAL-CHECKING-DFUN PCL::ACCESSOR-VALUES
            PCL::SET-CLASS-SLOT-VALUE-1
            PCL::GENERATE-DISCRIMINATION-NET
            PCL::REAL-MAKE-METHOD-LAMBDA
            PCL::ORDER-SPECIALIZERS WALKER::WALK-TEMPLATE
            PCL::ACCESSOR-MISS)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T T T) *)
            ITERATE::ITERATE-TRANSFORM-BODY)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T *) *) PCL::SLOT-VALUE-OR-DEFAULT
            PCL::MAKE-EFFECTIVE-METHOD-FUNCTION
            PCL::GET-EFFECTIVE-METHOD-FUNCTION
            PCL::MAKE-N-N-ACCESSOR-DFUN WALKER:NESTED-WALK-FORM
            PCL::MAKE-CHECKING-DFUN PCL::LOAD-DEFGENERIC
            PCL::TYPES-FROM-ARGUMENTS
            PCL::MAKE-DEFAULT-INITARGS-FORM-LIST
            PCL::MAKE-FINAL-ACCESSOR-DFUN PCL::MAKE-ACCESSOR-TABLE
            PCL::GET-SIMPLE-INITIALIZATION-FUNCTION
            PCL::GET-COMPLEX-INITIALIZATION-FUNCTIONS
            PCL::COMPUTE-SECONDARY-DISPATCH-FUNCTION)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T *) *)
            PCL::MAKE-EFFECTIVE-METHOD-FUNCTION-SIMPLE1
            ITERATE::RENAME-LET-BINDINGS)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T *) *) PCL::MAKE-ONE-INDEX-ACCESSOR-DFUN
            WALKER::WALK-DECLARATIONS
            PCL::GET-SECONDARY-DISPATCH-FUNCTION)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T T *) *) PCL::REAL-MAKE-A-METHOD)) 
(PROCLAIM '(FTYPE (FUNCTION (T STREAM T) T) PCL::PRINT-DFUN-INFO)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T) T) ITERATE::SIMPLE-EXPAND-GATHERING-FORM
            ITERATE::RENAME-AND-CAPTURE-VARIABLES
            ITERATE::VARIABLE-SAME-P
            PCL::GET-FUNCTION-GENERATOR
            WALKER:VARIABLE-DECLARATION PCL::GET-NEW-FUNCTION-GENERATOR
            PCL::TRACE-METHOD-INTERNAL PCL::ONE-INDEX-DFUN-INFO
            PCL::ONE-CLASS-DFUN-INFO
            PCL::MAP-ALL-ORDERS SYSTEM::APPLY-DISPLAY-FUN
            PCL::NOTE-PV-TABLE-REFERENCE
            WALKER::RELIST-INTERNAL
            PCL::MAKE-DFUN-CALL
            WALKER::WALK-TAGBODY-1 WALKER::WALK-LAMBDA
            PCL::OPTIMIZE-GF-CALL-INTERNAL
            PCL::SKIP-OPTIMIZE-SLOT-VALUE-BY-CLASS-P
            WALKER::WALK-COMPILER-LET PCL::SKIP-FAST-SLOT-ACCESS-P
            WALKER::WALK-UNEXPECTED-DECLARE WALKER::WALK-FLET
            WALKER::WALK-IF
            WALKER::WALK-LABELS WALKER::WALK-LET WALKER::WALK-LET*
            WALKER::WALK-LOCALLY
            WALKER::WALK-MACROLET
            PCL::FIX-SLOT-ACCESSORS WALKER::WALK-MULTIPLE-VALUE-BIND
            PCL:COMPUTE-EFFECTIVE-METHOD WALKER::WALK-SETQ
            WALKER::WALK-SYMBOL-MACROLET PCL::EMIT-SLOT-READ-FORM
            WALKER::WALK-TAGBODY PCL::EMIT-BOUNDP-CHECK WALKER::WALK-DO
            WALKER::WALK-DO* WALKER::WALK-PROG
            WALKER::WALK-NAMED-LAMBDA WALKER::WALK-PROG*
            PCL::EXPAND-DEFGENERIC PCL::EMIT-GREATER-THAN-1-DLAP
            PCL::EMIT-1-T-DLAP
            PCL::MAKE-METHOD-INITARGS-FORM-INTERNAL
            PCL::ENTRY-IN-CACHE-P PCL::CONVERT-TABLE
            PCL::MAKE-METHOD-SPEC PCL::TRACE-EMF-CALL-INTERNAL
            PCL::FLUSH-CACHE-TRAP PCL::SET-FUNCTION-NAME-1
            PCL::OBSOLETE-INSTANCE-TRAP
            PCL::COMPUTE-PRECEDENCE PCL::PRINT-STD-INSTANCE
            PCL::|SETF PCL METHOD-FUNCTION-GET|
            PCL::|SETF PCL PLIST-VALUE|
            WALKER::WITH-AUGMENTED-ENVIRONMENT-INTERNAL
            PCL::CAN-OPTIMIZE-ACCESS PCL::OPTIMIZE-SLOT-VALUE
            PCL::OPTIMIZE-SET-SLOT-VALUE PCL::DECLARE-STRUCTURE
            PCL::OPTIMIZE-SLOT-BOUNDP
            PCL::PRINT-CACHE PCL::COMPUTE-STD-CPL-PHASE-3
            PCL::FIRST-FORM-TO-LISP
            ITERATE::OPTIMIZE-ITERATE-FORM
            PCL::WRAP-METHOD-GROUP-SPECIFIER-BINDINGS
            PCL::MAKE-TOP-LEVEL-FORM PCL::INVALIDATE-WRAPPER
            PCL::STANDARD-COMPUTE-EFFECTIVE-METHOD
            PCL::MAKE-OPTIMIZED-STD-SLOT-VALUE-USING-CLASS-METHOD-FUNCTION
            PCL::MAKE-OPTIMIZED-STD-SETF-SLOT-VALUE-USING-CLASS-METHOD-FUNCTION
            PCL::MAKE-OPTIMIZED-STD-SLOT-BOUNDP-USING-CLASS-METHOD-FUNCTION
            WALKER::RECONS ITERATE::OPTIMIZE-GATHERING-FORM)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T) T) PCL::MAKE-EFFECTIVE-METHOD-FUNCTION1
            PCL::MAKE-EFFECTIVE-METHOD-FUNCTION-INTERNAL
            PCL::MAKE-EFFECTIVE-METHOD-FUNCTION-TYPE
            PCL::MEMF-TEST-CONVERTER
            PCL::LOAD-PRECOMPILED-DFUN-CONSTRUCTOR
            PCL::TWO-CLASS-DFUN-INFO
            WALKER::WALK-LET/LET* WALKER::WALK-PROG/PROG*
            WALKER::WALK-DO/DO*
            WALKER::WALK-BINDINGS-2 PCL::OPTIMIZE-READER
            PCL::OPTIMIZE-WRITER
            PCL::EMIT-CHECKING-OR-CACHING-FUNCTION-PRELIMINARY
            PCL::MAYBE-EXPAND-ACCESSOR-FORM
            PCL::INITIALIZE-INSTANCE-SIMPLE
            PCL::GET-WRAPPERS-FROM-CLASSES
            PCL::LOAD-PRECOMPILED-IIS-ENTRY
            PCL::FILL-CACHE-P
            PCL::ADJUST-CACHE
            PCL::EXPAND-CACHE
            PCL::EXPAND-SYMBOL-MACROLET-INTERNAL
            PCL::BOOTSTRAP-SET-SLOT PCL::EXPAND-DEFCLASS PCL::GET-CACHE
            )) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T) T) PCL::LOAD-FUNCTION-GENERATOR
            PCL::EXPAND-EMF-CALL-METHOD PCL::MAKE-FGEN
            PCL::BOOTSTRAP-MAKE-SLOT-DEFINITIONS
            PCL::BOOTSTRAP-ACCESSOR-DEFINITIONS1
            PCL::MAKE-FINAL-ORDINARY-DFUN-INTERNAL PCL::COMPUTE-PV-SLOT
            WALKER::WALK-BINDINGS-1
            PCL::OPTIMIZE-INSTANCE-ACCESS
            PCL::OPTIMIZE-ACCESSOR-CALL
            PCL::MAKE-METHOD-INITARGS-FORM-INTERNAL1
            PCL::UPDATE-SLOTS-IN-PV
            PCL::MAKE-PARAMETER-REFERENCES
            PCL::MAKE-EMF-CACHE
            PCL::GET-MAKE-INSTANCE-FUNCTION-INTERNAL
            PCL::MAKE-INSTANCE-FUNCTION-COMPLEX
            PCL::MAKE-INSTANCE-FUNCTION-SIMPLE
            PCL::OPTIMIZE-GENERIC-FUNCTION-CALL
            PCL::REAL-MAKE-METHOD-INITARGS-FORM
            )) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T *) T)
            PCL::MAKE-EFFECTIVE-METHOD-FUNCTION-SIMPLE
            PCL::MAKE-EMF-FROM-METHOD
            PCL::EXPAND-EFFECTIVE-METHOD-FUNCTION
            PCL::NAMED-OBJECT-PRINT-FUNCTION PCL::FIND-CLASS-FROM-CELL
            PCL::FIND-CLASS-PREDICATE-FROM-CELL PCL::INITIALIZE-INFO
            PCL::GET-EFFECTIVE-METHOD-FUNCTION1 PCL::GET-DECLARATION
            PCL::GET-METHOD-FUNCTION-PV-CELL
            PCL:ENSURE-GENERIC-FUNCTION-USING-CLASS PCL::EMIT-MISS
            PCL::METHOD-FUNCTION-GET PCL::PROBE-CACHE PCL::MAP-CACHE
            PCL::GET-CACHE-FROM-CACHE PCL::PRECOMPUTE-EFFECTIVE-METHODS
            PCL::RECORD-DEFINITION WALKER::CONVERT-MACRO-TO-LAMBDA
            PCL::CPL-ERROR PCL::REAL-ADD-METHOD
            PCL::REAL-ENSURE-GF-USING-CLASS--GENERIC-FUNCTION
            PCL::REAL-ENSURE-GF-USING-CLASS--NULL
            PCL::COMPUTE-SECONDARY-DISPATCH-FUNCTION1)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T T T T) T)
            PCL::GET-SECONDARY-DISPATCH-FUNCTION2)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T T) T)
            PCL::BOOTSTRAP-MAKE-SLOT-DEFINITION PCL::EMIT-SLOT-ACCESS
            PCL::OPTIMIZE-GF-CALL PCL::SET-ARG-INFO1 PCL::LOAD-DEFCLASS
            PCL::MAKE-EARLY-CLASS-DEFINITION)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T *) T) PCL::FILL-DFUN-CACHE
            PCL::EARLY-ADD-NAMED-METHOD PCL::REAL-ADD-NAMED-METHOD)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T T *) T) PCL::LOAD-DEFMETHOD
            PCL::MAKE-DEFMETHOD-FORM PCL::MAKE-DEFMETHOD-FORM-INTERNAL
            PCL::EARLY-MAKE-A-METHOD)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T *) T) PCL::CHECK-INITARGS-2-PLIST
            PCL::CHECK-INITARGS-2-LIST WALKER::WALK-ARGLIST
            PCL::MAKE-EMF-CALL PCL::CAN-OPTIMIZE-ACCESS1
            PCL::EMIT-FETCH-WRAPPER PCL::FILL-CACHE
            PCL::REAL-GET-METHOD PCL::CHECK-INITARGS-1 PCL::GET-METHOD)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T *) T) PCL::EMIT-DLAP
            PCL::GET-SECONDARY-DISPATCH-FUNCTION1)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T FIXNUM) T) PCL::FILL-CACHE-FROM-CACHE-P)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T T T) T) PCL::EXPAND-DEFMETHOD
            PCL::LOAD-DEFMETHOD-INTERNAL
            )) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T T T T T T T *) T)
            PCL::BOOTSTRAP-INITIALIZE-CLASS)) 
(PROCLAIM
    '(FTYPE (FUNCTION NIL *) PCL::COUNT-ALL-DFUNS PCL::RENEW-SYS-FILES
            PCL::EMIT-N-N-READERS PCL::EMIT-N-N-WRITERS)) 
(PROCLAIM
    '(FTYPE (FUNCTION NIL T) PCL::GET-EFFECTIVE-METHOD-GENSYM
            PCL::SHOW-EMF-CALL-TRACE PCL::BOOTSTRAP-META-BRAID
            PCL::BOOTSTRAP-BUILT-IN-CLASSES PCL::LIST-ALL-DFUNS
            PCL::DEFAULT-METHOD-ONLY-DFUN-INFO
            PCL::INITIALIZE-CHECKING-OR-CACHING-FUNCTION-LIST
            PCL::CACHES-TO-ALLOCATE PCL::UPDATE-DISPATCH-DFUNS
            PCL::MAKE-CACHE PCL::RESET-PCL-PACKAGE
            PCL::IN-THE-COMPILER-P PCL::STRUCTURE-FUNCTIONS-EXIST-P
            PCL::ALLOCATE-FUNCALLABLE-INSTANCE-2
            PCL::%%ALLOCATE-INSTANCE--CLASS
            PCL::ALLOCATE-FUNCALLABLE-INSTANCE-1
            PCL::DISPATCH-DFUN-INFO PCL::INITIAL-DISPATCH-DFUN-INFO
            PCL::INITIAL-DFUN-INFO PCL::NO-METHODS-DFUN-INFO
            PCL::SHOW-FREE-CACHE-VECTORS PCL::MAKE-CPD
            PCL::MAKE-ARG-INFO PCL::SHOW-DFUN-CONSTRUCTORS)) 
(PROCLAIM
    '(FTYPE (FUNCTION (*) *) PCL::UNTRACE-METHOD
            PCL:INVALID-METHOD-ERROR PCL:METHOD-COMBINATION-ERROR
            PCL::LIST-LARGE-CACHES
            PCL::UPDATE-MAKE-INSTANCE-FUNCTION-TABLE)) 
(PROCLAIM '(FTYPE (FUNCTION (FIXNUM T *) *) PCL::FIND-FREE-CACHE-LINE)) 
(PROCLAIM
    '(FTYPE (FUNCTION (FIXNUM T T) *) PCL::COMPUTE-CACHE-PARAMETERS)) 
(PROCLAIM
    '(FTYPE (FUNCTION (*) T) PCL::|__si::MAKE-DFUN-INFO|
            PCL::|__si::MAKE-NO-METHODS| PCL::|__si::MAKE-INITIAL|
            PCL::|__si::MAKE-INITIAL-DISPATCH|
            PCL::|__si::MAKE-DISPATCH|
            PCL::|__si::MAKE-DEFAULT-METHOD-ONLY|
            PCL::|__si::MAKE-ACCESSOR-DFUN-INFO|
            PCL::|__si::MAKE-ONE-INDEX-DFUN-INFO|
            PCL::MAKE-FAST-METHOD-CALL PCL::|__si::MAKE-N-N|
            PCL::MAKE-FAST-INSTANCE-BOUNDP PCL::|__si::MAKE-ONE-CLASS|
            PCL::|__si::MAKE-TWO-CLASS| PCL::|__si::MAKE-ONE-INDEX|
            PCL::|__si::MAKE-CHECKING| PCL::|__si::MAKE-ARG-INFO|
            PCL::FIX-EARLY-GENERIC-FUNCTIONS PCL::STRING-APPEND
            PCL::|__si::MAKE-CACHING| PCL::|__si::MAKE-CONSTANT-VALUE|
            PCL::FALSE PCL::|STRUCTURE-OBJECT class constructor|
            PCL::PV-WRAPPERS-FROM-PV-ARGS PCL::MAKE-PV-TABLE
            PCL::|__si::MAKE-PV-TABLE| PCL::INTERN-PV-TABLE
            PCL::CALLED-FIN-WITHOUT-FUNCTION
            PCL::|__si::MAKE-STD-INSTANCE| PCL::TRUE
            PCL::MAKE-INITIALIZE-INFO PCL::|__si::MAKE-CACHE|
            PCL::MAKE-PROGN WALKER::UNBOUND-LEXICAL-FUNCTION
            PCL::|__si::MAKE-CLASS-PRECEDENCE-DESCRIPTION|
            PCL::MAKE-METHOD-CALL)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T) *) PCL::TYPE-FROM-SPECIALIZER
            PCL::*NORMALIZE-TYPE PCL::UNPARSE-TYPE
            PCL::DEFAULT-CODE-CONVERTER PCL::CONVERT-TO-SYSTEM-TYPE
            PCL::EMIT-CONSTANT-VALUE PCL::SFUN-P PCL::PCL-DESCRIBE
            PCL::GET-GENERIC-FUNCTION-INFO PCL::EARLY-METHOD-FUNCTION
            PCL::EARLY-METHOD-STANDARD-ACCESSOR-SLOT-NAME
            PCL::SPECIALIZER-FROM-TYPE PCL::CLASS-EQ-TYPE
            COMPILER::CAN-USE-PRINT-CIRCLE-P PCL::STRUCTURE-WRAPPER
            PCL::FIND-STRUCTURE-CLASS PCL::MAKE-DISPATCH-DFUN
            PCL::FIND-WRAPPER PCL::PARSE-DEFMETHOD
            PCL::PROTOTYPES-FOR-MAKE-METHOD-LAMBDA
            PCL::FORCE-CACHE-FLUSHES PCL::EMIT-ONE-CLASS-READER
            PCL::EMIT-ONE-CLASS-WRITER PCL::EMIT-TWO-CLASS-READER
            PCL::EMIT-TWO-CLASS-WRITER PCL::EMIT-ONE-INDEX-READERS
            PCL::EMIT-ONE-INDEX-WRITERS PCL::NET-CODE-CONVERTER
            PCL::EMIT-IN-CHECKING-CACHE-P PCL::COMPILE-IIS-FUNCTIONS
            PCL::ANALYZE-LAMBDA-LIST
            PCL::COMPUTE-APPLICABLE-METHODS-EMF
            PCL::GET-DISPATCH-FUNCTION PCL::INSURE-CACHING-DFUN
            PCL::%FBOUNDP PCL::CCLOSUREP PCL::GENERIC-FUNCTION-NAME-P
            PCL::MAKE-FINAL-DISPATCH-DFUN
            PCL::STRUCTURE-SLOTD-INIT-FORM
            PCL::PARSE-METHOD-GROUP-SPECIFIER
            PCL::METHOD-PROTOTYPE-FOR-GF
            PCL::EARLY-COLLECT-INHERITANCE)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T) T) PCL::UNENCAPSULATED-FDEFINITION
            PCL::DFUN-INFO-P PCL::NO-METHODS-P
            PCL::MAKE-TYPE-PREDICATE
            PCL::DEFAULT-TEST-CONVERTER PCL::INITIAL-P
            PCL::UNPARSE-TYPE-LIST PCL::MAKE-CALL-METHODS
            PCL::DEFAULT-CONSTANT-CONVERTER PCL::INITIAL-DISPATCH-P
            PCL::DISPATCH-P PCL::GBOUNDP PCL::GMAKUNBOUND
            PCL::DEFAULT-CONSTANTP PCL::DEFAULT-METHOD-ONLY-P
            PCL::FGEN-TEST PCL::LOOKUP-FGEN PCL::ACCESSOR-DFUN-INFO-P
            PCL::FGEN-GENERATOR PCL::FGEN-SYSTEM
            PCL::ONE-INDEX-DFUN-INFO-P PCL::FAST-METHOD-CALL-P
            PCL::N-N-P PCL::FAST-INSTANCE-BOUNDP-P
            PCL::METHOD-FUNCTION-PV-TABLE PCL::METHOD-FUNCTION-METHOD
            PCL::STORE-FGEN PCL::ONE-CLASS-P
            PCL::METHOD-FUNCTION-NEEDS-NEXT-METHODS-P
            PCL::FTYPE-DECLARATION-FROM-LAMBDA-LIST PCL::FGEN-GENSYMS
            PCL::TWO-CLASS-P PCL::ARG-INFO-LAMBDA-LIST
            PCL::ARG-INFO-PRECEDENCE PCL::ARG-INFO-METATYPES
            PCL::FGEN-GENERATOR-LAMBDA SYSTEM:%STRUCTURE-NAME
            PCL::ARG-INFO-NUMBER-OPTIONAL
            SYSTEM:%COMPILED-FUNCTION-NAME PCL::ARG-INFO-KEY/REST-P
            PCL::ONE-INDEX-P PCL::ARG-INFO-KEYWORDS
            PCL::GF-INFO-SIMPLE-ACCESSOR-TYPE
            PCL::GF-PRECOMPUTE-DFUN-AND-EMF-P
            PCL::GF-INFO-STATIC-C-A-M-EMF PCL::CHECKING-P
            PCL::GF-INFO-C-A-M-EMF-STD-P PCL::GF-INFO-FAST-MF-P
            PCL::UNDEFMETHOD-1 PCL::ARG-INFO-P
            PCL::FAST-METHOD-CALL-ARG-INFO PCL::ARG-INFO-NKEYS
            PCL::GF-DFUN-CACHE PCL:CLASS-OF PCL::GF-DFUN-INFO
            PCL::FUNCTION-RETURNING-NIL
            PCL::ACCESSOR-DFUN-INFO-ACCESSOR-TYPE PCL::EVAL-FORM
            PCL::ONE-INDEX-DFUN-INFO-INDEX
            PCL::SLOT-INITARGS-FROM-STRUCTURE-SLOTD PCL::TYPE-CLASS
            PCL::ONE-CLASS-WRAPPER0 PCL::EXTRACT-PARAMETERS
            PCL::CLASS-PREDICATE PCL::EXTRACT-REQUIRED-PARAMETERS
            PCL::MAKE-CLASS-EQ-PREDICATE PCL::TWO-CLASS-WRAPPER1
            PCL::MAKE-EQL-PREDICATE PCL::CHECKING-FUNCTION
            PCL::BOOTSTRAP-ACCESSOR-DEFINITIONS
            PCL::INITIALIZE-INFO-KEY PCL::BOOTSTRAP-CLASS-PREDICATES
            PCL::GET-BUILT-IN-CLASS-SYMBOL PCL::INITIALIZE-INFO-WRAPPER
            PCL::GET-BUILT-IN-WRAPPER-SYMBOL PCL::DO-STANDARD-DEFSETF-1
            PCL::CACHING-P PCL::GFS-OF-TYPE PCL::LEGAL-CLASS-NAME-P
            PCL::STRUCTURE-TYPE-P PCL::CONSTANT-VALUE-P
            PCL::USE-DEFAULT-METHOD-ONLY-DFUN-P
            SYSTEM::NEXT-STACK-FRAME PCL::WRAPPER-FIELD
            PCL::NEXT-WRAPPER-FIELD PCL::SETFBOUNDP
            PCL::GET-SETF-FUNCTION-NAME PCL::USE-CACHING-DFUN-P
            PCL::MAKE-PV-TYPE-DECLARATION
            PCL::MAKE-CALLS-TYPE-DECLARATION PCL::MAP-SPECIALIZERS
            WALKER:VARIABLE-GLOBALLY-SPECIAL-P PCL::SLOT-VECTOR-SYMBOL
            PCL::MAKE-PERMUTATION-VECTOR PCL::STRUCTURE-OBJECT-P
            PCL::EXPAND-MAKE-INSTANCE-FORM PCL::MAKE-CONSTANT-FUNCTION
            PCL::FUNCTION-RETURNING-T PCL::SORT-SLOTS PCL::SORT-CALLS
            PCL::SYMBOL-PKG-NAME
            PCL::CLASS-HAS-A-FORWARD-REFERENCED-SUPERCLASS-P
            PCL::INITIALIZE-INFO-BOUND-SLOTS
            PCL::INITIALIZE-INFO-CACHED-VALID-P
            PCL::GET-MAKE-INSTANCE-FUNCTIONS
            PCL::INITIALIZE-INFO-CACHED-RI-VALID-P
            PCL::INITIALIZE-INFO-CACHED-INITARGS-FORM-LIST
            PCL::INITIALIZE-INFO-CACHED-NEW-KEYS
            PCL::UPDATE-C-A-M-GF-INFO
            PCL::INITIALIZE-INFO-CACHED-DEFAULT-INITARGS-FUNCTION
            PCL::UPDATE-GF-SIMPLE-ACCESSOR-TYPE
            PCL::UPDATE-GFS-OF-CLASS
            PCL::INITIALIZE-INFO-CACHED-SHARED-INITIALIZE-T-FUNCTION
            PCL::DO-STANDARD-DEFSETFS-FOR-DEFCLASS
            PCL::STANDARD-SVUC-METHOD
            PCL::INITIALIZE-INFO-CACHED-SHARED-INITIALIZE-NIL-FUNCTION
            PCL:EXTRACT-LAMBDA-LIST PCL::%CCLOSURE-ENV
            PCL::STRUCTURE-SVUC-METHOD
            PCL::INITIALIZE-INFO-CACHED-CONSTANTS
            PCL:EXTRACT-SPECIALIZER-NAMES PCL::METHOD-FUNCTION-PLIST
            PCL::INITIALIZE-INFO-CACHED-COMBINED-INITIALIZE-FUNCTION
            PCL::INITIALIZE-INFO-CACHED-MAKE-INSTANCE-FUNCTION
            PCL::INITIALIZE-INFO-CACHED-MAKE-INSTANCE-FUNCTION-SYMBOL
            PCL::INTERNED-SYMBOL-P PCL::GDEFINITION
            PCL::UPDATE-CLASS-CAN-PRECEDE-P PCL::%STD-INSTANCE-WRAPPER
            PCL::%STD-INSTANCE-SLOTS PCL::PV-TABLEP PCL::STD-INSTANCE-P
            PCL::COMPUTE-MCASE-PARAMETERS PCL::COMPUTE-CLASS-SLOTS
            PCL::MAKE-PV-TABLE-TYPE-DECLARATION PCL::NET-TEST-CONVERTER
            PCL:INTERN-EQL-SPECIALIZER
            PCL::MAKE-INSTANCE-FUNCTION-SYMBOL
            PCL::UPDATE-ALL-C-A-M-GF-INFO
            PCL::UPDATE-PV-TABLE-CACHE-INFO PCL::DFUN-INFO-CACHE
            PCL::NO-METHODS-CACHE PCL::ARG-INFO-APPLYP
            PCL::INITIAL-CACHE PCL::INITIAL-DISPATCH-CACHE
            PCL::CHECK-CACHE PCL::DISPATCH-CACHE PCL::CLASS-FROM-TYPE
            PCL::DEFAULT-METHOD-ONLY-CACHE PCL::DNET-METHODS-P
            PCL::ACCESSOR-DFUN-INFO-CACHE
            PCL::METHOD-FUNCTION-FROM-FAST-FUNCTION
            PCL::ONE-INDEX-DFUN-INFO-CACHE
            PCL::ONE-INDEX-DFUN-INFO-ACCESSOR-TYPE
            PCL::METHOD-CALL-CALL-METHOD-ARGS PCL::KEYWORD-SPEC-NAME
            PCL::N-N-CACHE PCL::GENERIC-CLOBBERS-FUNCTION
            PCL::N-N-ACCESSOR-TYPE PCL::FAST-METHOD-CALL-PV-CELL
            PCL::WRAPPER-FOR-STRUCTURE PCL::ONE-CLASS-CACHE
            PCL::FAST-METHOD-CALL-NEXT-METHOD-CALL
            PCL::ONE-CLASS-ACCESSOR-TYPE PCL::ONE-CLASS-INDEX
            PCL::BUILT-IN-WRAPPER-OF PCL::TWO-CLASS-CACHE
            PCL::BUILT-IN-OR-STRUCTURE-WRAPPER1
            PCL::TWO-CLASS-ACCESSOR-TYPE PCL::TWO-CLASS-INDEX
            PCL::ALLOCATE-CACHE-VECTOR PCL::TWO-CLASS-WRAPPER0
            PCL::FLUSH-CACHE-VECTOR-INTERNAL PCL::ONE-INDEX-CACHE
            PCL::EARLY-CLASS-NAME PCL::ONE-INDEX-ACCESSOR-TYPE
            PCL::ONE-INDEX-INDEX PCL::INTERN-FUNCTION-NAME
            PCL::CHECKING-CACHE PCL::COMPILE-LAMBDA-UNCOMPILED
            PCL::GF-LAMBDA-LIST PCL::CACHING-CACHE
            PCL::CONSTANT-VALUE-CACHE PCL::COMPILE-LAMBDA-DEFERRED
            PCL::FUNCALLABLE-INSTANCE-P
            PCL::RESET-CLASS-INITIALIZE-INFO PCL::GET-CACHE-VECTOR
            PCL::CONSTANT-SYMBOL-P PCL::FREE-CACHE-VECTOR
            PCL::EARLY-METHOD-LAMBDA-LIST PCL::ARG-INFO-VALID-P
            PCL::DFUN-ARG-SYMBOL PCL::EARLY-METHOD-CLASS
            PCL::EARLY-GF-P PCL::EARLY-GF-NAME PCL::CACHING-DFUN-INFO
            PCL::COMPUTE-APPLICABLE-METHODS-EMF-STD-P
            PCL::CONSTANT-VALUE-DFUN-INFO
            PCL::RESET-CLASS-INITIALIZE-INFO-1 PCL::FREE-CACHE
            PCL::PARSE-SPECIALIZERS PCL::RESET-INITIALIZE-INFO
            PCL::EARLY-METHOD-QUALIFIERS
            PCL::PROCLAIM-INCOMPATIBLE-SUPERCLASSES PCL::WRAPPER-OF
            PCL::EARLY-METHOD-STANDARD-ACCESSOR-P
            PCL::FUNCTION-PRETTY-ARGLIST
            PCL::GET-MAKE-INSTANCE-FUNCTION PCL::CHECK-WRAPPER-VALIDITY
            PCL::UNPARSE-SPECIALIZERS PCL::%SYMBOL-FUNCTION
            PCL::FINAL-ACCESSOR-DFUN-TYPE
            PCL::COMPLICATED-INSTANCE-CREATION-METHOD
            PCL::DEFAULT-STRUCTUREP PCL::UPDATE-GF-INFO
            PCL::CACHE-OWNER PCL::DEFAULT-STRUCTURE-INSTANCE-P
            PCL::DEFAULT-STRUCTURE-TYPE PCL::STRUCTURE-TYPE
            PCL::COMPUTE-STD-CPL-PHASE-2 PCL::GET-PV-CELL-FOR-CLASS
            PCL::STRUCTURE-TYPE-INCLUDED-TYPE-NAME
            PCL::STRUCTURE-TYPE-SLOT-DESCRIPTION-LIST PCL::CACHE-P
            PCL::STRUCTURE-SLOTD-NAME
            PCL::STRUCTURE-SLOTD-ACCESSOR-SYMBOL
            PCL::DEFAULT-SECONDARY-DISPATCH-FUNCTION
            PCL::STRUCTURE-SLOTD-WRITER-FUNCTION
            PCL::FIND-CYCLE-REASONS PCL::EARLY-CLASS-DEFINITION
            PCL::ECD-SOURCE PCL::STRUCTURE-SLOTD-TYPE
            PCL::FORMAT-CYCLE-REASONS PCL::ECD-METACLASS PCL::CPD-CLASS
            PCL::EARLY-CLASS-PRECEDENCE-LIST
            PCL::METHODS-CONTAIN-EQL-SPECIALIZER-P PCL::CPD-SUPERS
            PCL::EXPAND-LONG-DEFCOMBIN PCL::EARLY-CLASS-NAME-OF
            PCL::CPD-AFTER PCL::EXPAND-SHORT-DEFCOMBIN
            PCL::EARLY-CLASS-SLOTDS PCL::CPD-COUNT
            PCL::EARLY-SLOT-DEFINITION-NAME PCL::SLOT-READER-SYMBOL
            PCL::EARLY-SLOT-DEFINITION-LOCATION WALKER::ENV-LOCK
            PCL::MAKE-INITIAL-DFUN PCL::EARLY-ACCESSOR-METHOD-SLOT-NAME
            PCL::SLOT-WRITER-SYMBOL WALKER::ENV-DECLARATIONS
            WALKER::ENV-LEXICAL-VARIABLES PCL::LIST-DFUN
            PCL::SLOT-BOUNDP-SYMBOL PCL::MAP-ALL-GENERIC-FUNCTIONS
            PCL::MAKE-STRUCTURE-SLOT-BOUNDP-FUNCTION
            PCL::EARLY-CLASS-DIRECT-SUBCLASSES
            PCL::MAKE-FUNCTION-INLINE PCL::LIST-LARGE-CACHE
            PCL::CLASS-PRECEDENCE-DESCRIPTION-P
            PCL::INFORM-TYPE-SYSTEM-ABOUT-STD-CLASS
            PCL::MAKE-DEFAULT-METHOD-GROUP-DESCRIPTION
            PCL::MAKE-OPTIMIZED-STRUCTURE-SLOT-VALUE-USING-CLASS-METHOD-FUNCTION
            WALKER::ENV-WALK-FUNCTION
            WALKER::GET-IMPLEMENTATION-DEPENDENT-WALKER-TEMPLATE
            PCL::COUNT-DFUN PCL::MAKE-INITFUNCTION
            PCL::MAKE-OPTIMIZED-STRUCTURE-SETF-SLOT-VALUE-USING-CLASS-METHOD-FUNCTION
            ITERATE::VARIABLES-FROM-LET WALKER::ENV-WALK-FORM
            PCL::MAKE-OPTIMIZED-STRUCTURE-SLOT-BOUNDP-USING-CLASS-METHOD-FUNCTION
            PCL::INITIALIZE-INFO-P PCL::ECD-CLASS-NAME PCL::COPY-CACHE
            PCL::COMPUTE-LINE-SIZE PCL::CANONICAL-SLOT-NAME
            WALKER::GET-WALKER-TEMPLATE PCL::EARLY-CLASS-SLOTS
            PCL::STRUCTURE-TYPE-INTERNAL-SLOTDS PCL::EARLY-COLLECT-CPL
            PCL::EARLY-COLLECT-SLOTS
            PCL::METHOD-LL->GENERIC-FUNCTION-LL
            PCL::EARLY-COLLECT-DEFAULT-INITARGS
            PCL::ECD-SUPERCLASS-NAMES PCL::METHOD-CALL-P
            PCL::STRUCTURE-SLOT-BOUNDP ITERATE::SEQUENCE-ACCESSOR
            PCL::ECD-CANONICAL-SLOTS PCL::ECD-OTHER-INITARGS)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T *) *) PCL::COERCE-TO-CLASS
            PCL::GET-METHOD-FUNCTION WALKER:MACROEXPAND-ALL
            PCL::GET-FUNCTION PCL::GET-FUNCTION1
            PCL:ENSURE-GENERIC-FUNCTION PCL::PARSE-METHOD-OR-SPEC
            PCL::EXTRACT-DECLARATIONS PCL::GET-DFUN-CONSTRUCTOR
            PCL::MAP-ALL-CLASSES PCL::MAKE-CACHING-DFUN
            WALKER:WALK-FORM PCL:ENSURE-CLASS
            PCL::MAKE-METHOD-FUNCTION-INTERNAL
            PCL::PARSE-SPECIALIZED-LAMBDA-LIST
            PCL::MAKE-METHOD-LAMBDA-INTERNAL
            PCL::MAKE-CONSTANT-VALUE-DFUN PCL::MAKE-FINAL-DFUN-INTERNAL
            PCL::COMPILE-LAMBDA)) 
(PROCLAIM '(FTYPE (FUNCTION (T T *) (VALUES T T)) PCL::SYMBOL-APPEND)) 
(PROCLAIM '(FTYPE (FUNCTION (T *) STRING) PCL::CAPITALIZE-WORDS)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T) *) PCL::SAUT-CLASS
            PCL::SPECIALIZER-APPLICABLE-USING-TYPE-P PCL::*TYPEP
            PCL::COMPUTE-TEST PCL::GET-NEW-FUNCTION-GENERATOR-INTERNAL
            PCL::COMPUTE-CODE PCL::CLASS-APPLICABLE-USING-CLASS-P
            PCL::SAUT-AND PCL::SAUT-NOT PCL::SAUT-PROTOTYPE
            COMPILER::CAN-USE-PRINT-CIRCLE-P1 PCL:SLOT-BOUNDP
            PCL::DESTRUCTURE PCL:SLOT-MAKUNBOUND PCL:SLOT-VALUE
            PCL::ENSURE-CLASS-VALUES PCL::MAKE-DIRECT-SLOTD
            PCL::GENERATE-FAST-CLASS-SLOT-ACCESS-P
            PCL::MUTATE-SLOTS-AND-CALLS PCL::INVOKE-EMF
            PCL::EMIT-DEFAULT-ONLY-FUNCTION PCL::SPLIT-DECLARATIONS
            PCL::EMIT-DEFAULT-ONLY COMPILER::C2LAMBDA-EXPR-WITH-KEY
            PCL::SLOT-NAME-LISTS-FROM-SLOTS PCL::EMIT-CHECKING
            PCL::UPDATE-SLOT-VALUE-GF-INFO PCL::EMIT-CACHING
            PCL::SDFUN-FOR-CACHING PCL::SLOT-UNBOUND-INTERNAL
            PCL::MAKE-INSTANCE-1 PCL::SET-FUNCTION-NAME
            PCL::COMPUTE-STD-CPL-PHASE-1 PCL::FORM-LIST-TO-LISP
            PCL::FIND-SUPERCLASS-CHAIN PCL::SAUT-CLASS-EQ
            PCL::COMPUTE-APPLICABLE-METHODS-USING-TYPES
            PCL::CHECK-INITARGS-VALUES PCL::SAUT-EQL PCL::*SUBTYPEP
            ITERATE::PARSE-DECLARATIONS PCL::INITIAL-DFUN)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T *) T) PCL::MAKE-TYPE-PREDICATE-NAME
            PCL::SET-DFUN PCL:FIND-CLASS PCL::TRACE-METHOD
            PCL::FIND-CLASS-CELL PCL::MAKE-FINAL-DFUN
            PCL::PV-TABLE-LOOKUP-PV-ARGS PCL::USE-DISPATCH-DFUN-P
            WALKER::RELIST* WALKER::RELIST PCL::FIND-CLASS-PREDICATE
            PCL::EARLY-METHOD-SPECIALIZERS
            PCL::USE-CONSTANT-VALUE-DFUN-P PCL::MAKE-EARLY-GF
            PCL::ALLOCATE-FUNCALLABLE-INSTANCE PCL::SET-ARG-INFO
            PCL::INITIALIZE-METHOD-FUNCTION PCL::UPDATE-DFUN
            PCL::MAKE-SPECIALIZABLE PCL::ALLOCATE-STRUCTURE-INSTANCE
            PCL::ALLOCATE-STANDARD-INSTANCE
            WALKER::WALKER-ENVIRONMENT-BIND-1
            ITERATE::FUNCTION-LAMBDA-P ITERATE::MAYBE-WARN
            PCL::MAKE-WRAPPER)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T T) (*)) PCL::SORT-APPLICABLE-METHODS
            PCL::SORT-METHODS)) 
(PROCLAIM
    '(FTYPE (FUNCTION (T T) T) PCL::FDEFINE-CAREFULLY
            PCL::MAKE-INTERNAL-READER-METHOD-FUNCTION
            PCL::MAKE-STD-READER-METHOD-FUNCTION
            PCL::MAKE-STD-WRITER-METHOD-FUNCTION
            ITERATE::SIMPLE-EXPAND-ITERATE-FORM
            PCL::MAKE-STD-BOUNDP-METHOD-FUNCTION
            PCL::DO-SATISFIES-DEFTYPE PCL::MEMF-CONSTANT-CONVERTER
            PCL::COMPUTE-CONSTANTS PCL::CLASS-CAN-PRECEDE-P
            PCL::SAUT-NOT-CLASS PCL::SAUT-NOT-CLASS-EQ
            PCL::SAUT-NOT-PROTOTYPE PCL::GF-MAKE-FUNCTION-FROM-EMF
            PCL::SAUT-NOT-EQL PCL::SUPERCLASSES-COMPATIBLE-P
            PCL::CLASSES-HAVE-COMMON-SUBCLASS-P
            SYSTEM:%SET-COMPILED-FUNCTION-NAME PCL:ADD-METHOD
            SYSTEM::DISPLAY-ENV PCL::DESCRIBE-PACKAGE
            SYSTEM::DISPLAY-COMPILED-ENV
            PCL::PRINTING-RANDOM-THING-INTERNAL
            PCL::MAKE-CLASS-PREDICATE
            PCL::METHOD-FUNCTION-RETURNING-NIL
            PCL::METHOD-FUNCTION-RETURNING-T PCL::VARIABLE-CLASS
            PCL::MAKE-PLIST PCL::REMTAIL PCL:REMOVE-METHOD
            PCL:SLOT-EXISTS-P PCL::DESTRUCTURE-INTERNAL
            PCL::ACCESSOR-MISS-FUNCTION
            PCL::UPDATE-INITIALIZE-INFO-INTERNAL PCL::N-N-DFUN-INFO
            PCL::MAKE-CAXR PCL::MAKE-CDXR WALKER:VARIABLE-LEXICAL-P
            WALKER:VARIABLE-SPECIAL-P PCL::CHECKING-DFUN-INFO
            PCL::MAKE-PV-TABLE-INTERNAL PCL::FIND-SLOT-DEFINITION
            WALKER::WALK-REPEAT-EVAL WALKER::NOTE-DECLARATION
            PCL::MAKE-DFUN-LAMBDA-LIST WALKER::NOTE-LEXICAL-BINDING
            PCL::MAKE-DLAP-LAMBDA-LIST PCL::ADD-DIRECT-SUBCLASSES
            PCL::COMPUTE-PV PCL::MAKE-DFUN-ARG-LIST PCL::COMPUTE-CALLS
            PCL::MAKE-FAST-METHOD-CALL-LAMBDA-LIST
            PCL::UPDATE-ALL-PV-TABLE-CACHES PCL::UPDATE-CLASS
            PCL::MAP-PV-TABLE-REFERENCES-OF PCL::ADD-SLOT-ACCESSORS
            WALKER::ENVIRONMENT-FUNCTION PCL::REMOVE-DIRECT-SUBCLASSES
            PCL::REMOVE-SLOT-ACCESSORS PCL::SYMBOL-LESSP
            PCL::SYMBOL-OR-CONS-LESSP PCL::|SETF PCL FIND-CLASS|
            PCL::|SETF PCL FIND-CLASS-PREDICATE|
            PCL::PV-WRAPPERS-FROM-ALL-ARGS PCL::PV-TABLE-LOOKUP
            PCL::PROCLAIM-DEFGENERIC PCL::UPDATE-CPL PCL::LIST-EQ
            PCL::UPDATE-SLOTS PCL::COMPUTE-APPLICABLE-METHODS-FUNCTION
            PCL::COMPUTE-EMF-FROM-WRAPPERS PCL::UPDATE-INITS
            PCL::UPDATE-STD-OR-STR-METHODS
            PCL::SET-STANDARD-SVUC-METHOD PCL::EMIT-1-NIL-DLAP
            PCL::PLIST-VALUE PCL::SET-STRUCTURE-SVUC-METHOD
            PCL::EMIT-1-WRAPPER-COMPUTE-PRIMARY-CACHE-LOCATION
            PCL:FUNCALLABLE-STANDARD-INSTANCE-ACCESS
            PCL::MEC-ALL-CLASSES-INTERNAL
            PCL::EMIT-N-WRAPPER-COMPUTE-PRIMARY-CACHE-LOCATION
            PCL::MEC-ALL-CLASSES PCL::%SET-CCLOSURE-ENV
            PCL::MEC-ALL-CLASS-LISTS PCL::REDEFINE-FUNCTION
            PCL::METHODS-CONVERTER PCL::COMPUTE-LAYOUT PCL::NO-SLOT
            PCL::PV-WRAPPERS-FROM-ALL-WRAPPERS
            PCL::NET-CONSTANT-CONVERTER PCL::AUGMENT-TYPE
            PCL::CHANGE-CLASS-INTERNAL
            PCL:SET-FUNCALLABLE-INSTANCE-FUNCTION
            PCL::VALUE-FOR-CACHING PCL:STANDARD-INSTANCE-ACCESS
            PCL::|SETF PCL METHOD-FUNCTION-PLIST| PCL::GET-KEY-ARG
            PCL::GET-KEY-ARG1 PCL::SET-METHODS
            PCL::SET-FUNCTION-PRETTY-ARGLIST
            PCL::FIND-STANDARD-II-METHOD PCL::MAKE-EARLY-ACCESSOR
            PCL::DOCTOR-DFUN-FOR-THE-DEBUGGER PCL::COMPUTE-STD-CPL
            PCL::|SETF PCL GDEFINITION|
            PCL::MAKE-DISCRIMINATING-FUNCTION-ARGLIST PCL::ADD-FORMS
            PCL::CPL-INCONSISTENT-ERROR
            PCL::REDIRECT-EARLY-FUNCTION-INTERNAL PCL::ADD-TO-CVECTOR
            PCL::BOOTSTRAP-SLOT-INDEX PCL::QUALIFIER-CHECK-RUNTIME
            PCL::CPL-FORWARD-REFERENCED-CLASS-ERROR
            PCL::REAL-REMOVE-METHOD WALKER::ENVIRONMENT-MACRO
            PCL::CANONICALIZE-SLOT-SPECIFICATION
            PCL::CANONICALIZE-DEFCLASS-OPTION PCL::SET-WRAPPER
            PCL::DEAL-WITH-ARGUMENTS-OPTION
            PCL::PARSE-QUALIFIER-PATTERN PCL::SWAP-WRAPPERS-AND-SLOTS
            ITERATE::MV-SETQ PCL::MAKE-UNORDERED-METHODS-EMF
            PCL::CLASS-MIGHT-PRECEDE-P
            ITERATE::EXTRACT-SPECIAL-BINDINGS
            WALKER::VARIABLE-SYMBOL-MACRO-P PCL::RAISE-METATYPE)) 
(PROCLAIM '(FTYPE (FUNCTION NIL FIXNUM) PCL::GET-WRAPPER-CACHE-NUMBER)) 
(DOLIST (PCL::V '(PCL::ADD-READER-METHOD
                     PCL::SHORT-COMBINATION-IDENTITY-WITH-ONE-ARGUMENT
                     PCL::REMOVE-READER-METHOD PCL::EQL-SPECIALIZER-P
                     PCL::OBJECT-PLIST
                     PCL::SLOT-DEFINITION-DEFSTRUCT-ACCESSOR-SYMBOL
                     PCL::SPECIALIZER-TYPE PCL::GF-DFUN-STATE
                     PCL::CLASS-DEFSTRUCT-CONSTRUCTOR
                     PCL::METHOD-FAST-FUNCTION PCL::SPECIALIZERP
                     PCL::EXACT-CLASS-SPECIALIZER-P
                     PCL::COMPATIBLE-META-CLASS-CHANGE-P
                     PCL::UPDATE-GF-DFUN PCL::SPECIALIZER-OBJECT
                     PCL::ACCESSOR-METHOD-SLOT-NAME
                     PCL::SPECIALIZER-CLASS PCL::CLASS-EQ-SPECIALIZER-P
                     PCL::SLOTS-FETCHER PCL::REMOVE-WRITER-METHOD
                     PCL::STRUCTURE-CLASS-P 
                     PCL::DOCUMENTATION PCL::METHOD-PRETTY-ARGLIST
                     PCL::CLASS-EQ-SPECIALIZER
                     PCL::INFORM-TYPE-SYSTEM-ABOUT-CLASS
                     PCL::ACCESSOR-METHOD-CLASS
                     PCL::GENERIC-FUNCTION-PRETTY-ARGLIST
                     PCL::MAKE-BOUNDP-METHOD-FUNCTION
                     PCL::CLASS-PREDICATE-NAME PCL::CLASSP
                     PCL::LEGAL-QUALIFIERS-P PCL::ADD-BOUNDP-METHOD
                     PCL::LEGAL-LAMBDA-LIST-P
                     PCL::|SETF PCL GENERIC-FUNCTION-NAME|
                     PCL::DESCRIBE-OBJECT PCL::CLASS-INITIALIZE-INFO
                     PCL::MAKE-WRITER-METHOD-FUNCTION
                     PCL::|SETF PCL GF-DFUN-STATE|
                     PCL::|SETF PCL SLOT-DEFINITION-NAME|
                     PCL::|SETF PCL CLASS-NAME|
                     PCL::INITIALIZE-INTERNAL-SLOT-FUNCTIONS
                     PCL::|SETF PCL SLOT-DEFINITION-TYPE|
                     PCL::METHOD-COMBINATION-P
                     PCL::|SETF PCL GENERIC-FUNCTION-METHODS|
                     PCL::|SETF PCL GENERIC-FUNCTION-METHOD-COMBINATION|
                     PCL::|SETF PCL METHOD-GENERIC-FUNCTION|
                     PCL::|SETF PCL SLOT-ACCESSOR-STD-P|
                     PCL::LEGAL-SPECIALIZERS-P
                     PCL::|SETF PCL OBJECT-PLIST|
                     PCL::|SETF PCL SLOT-DEFINITION-INITFORM|
                     PCL::|SETF PCL CLASS-DEFSTRUCT-FORM|
                     PCL::|SETF PCL GENERIC-FUNCTION-METHOD-CLASS|
                     PCL::SLOT-ACCESSOR-STD-P
                     PCL::|SETF PCL GF-PRETTY-ARGLIST|
                     PCL::|SETF PCL SLOT-ACCESSOR-FUNCTION|
                     PCL::|SETF PCL SLOT-DEFINITION-LOCATION|
                     PCL::|SETF PCL SLOT-DEFINITION-READER-FUNCTION|
                     PCL::|SETF PCL SLOT-DEFINITION-WRITER-FUNCTION|
                     PCL::|SETF PCL SLOT-DEFINITION-BOUNDP-FUNCTION|
                     PCL::|SETF PCL SLOT-DEFINITION-INTERNAL-READER-FUNCTION|
                     PCL::|SETF PCL SLOT-DEFINITION-INTERNAL-WRITER-FUNCTION|
                     PCL::|SETF PCL SLOT-DEFINITION-ALLOCATION|
                     PCL::|SETF PCL SLOT-DEFINITION-INITFUNCTION|
                     PCL::METHOD-COMBINATION-OPTIONS
                     PCL::|SETF PCL SLOT-DEFINITION-READERS|
                     PCL::|SETF PCL DOCUMENTATION|
                     PCL::FUNCALLABLE-STANDARD-CLASS-P
                     PCL::|SETF PCL SLOT-DEFINITION-CLASS|
                     PCL::|SETF PCL SLOT-VALUE-USING-CLASS|
                     PCL::CLASS-CAN-PRECEDE-LIST
                     PCL::|SETF PCL CLASS-DIRECT-SLOTS|
                     PCL::|SETF PCL CLASS-SLOTS|
                     PCL::SLOT-ACCESSOR-FUNCTION
                     PCL::|SETF PCL CLASS-INCOMPATIBLE-SUPERCLASS-LIST|
                     PCL::|SETF PCL SLOT-DEFINITION-WRITERS|
                     PCL::SLOT-CLASS-P PCL::MAKE-READER-METHOD-FUNCTION
                     PCL::LEGAL-METHOD-FUNCTION-P PCL::GET-METHOD
                     PCL::SHORT-METHOD-COMBINATION-P PCL::GF-ARG-INFO
                     PCL::SPECIALIZER-METHOD-TABLE
                     PCL::MAKE-METHOD-INITARGS-FORM
                     PCL::CLASS-DEFSTRUCT-FORM PCL::GF-PRETTY-ARGLIST
                     PCL::SAME-SPECIALIZER-P
                     PCL::SLOT-DEFINITION-BOUNDP-FUNCTION
                     PCL::SLOT-DEFINITION-WRITER-FUNCTION
                     PCL::SLOT-DEFINITION-READER-FUNCTION
                     PCL::SLOT-DEFINITION-INTERNAL-WRITER-FUNCTION
                     PCL::SLOT-DEFINITION-INTERNAL-READER-FUNCTION
                     PCL::SLOT-DEFINITION-CLASS
                     PCL::EQL-SPECIALIZER-OBJECT
                     PCL::CLASS-CONSTRUCTORS PCL::SLOTS-TO-INSPECT
                     PCL::SPECIALIZER-DIRECT-GENERIC-FUNCTIONS
                     PCL::ADD-WRITER-METHOD
                     PCL::LONG-METHOD-COMBINATION-FUNCTION
                     PCL::GENERIC-FUNCTION-P PCL::LEGAL-SLOT-NAME-P
                     PCL::CLASS-WRAPPER PCL::DEFINITION-SOURCE
                     PCL::DEFAULT-INITARGS PCL::CLASS-SLOT-VALUE
                     PCL::FORWARD-REFERENCED-CLASS-P
                     PCL::GF-FAST-METHOD-FUNCTION-P
                     PCL::LEGAL-QUALIFIER-P PCL::METHOD-P
                     PCL::CLASS-SLOT-CELLS
                     PCL::STANDARD-ACCESSOR-METHOD-P
                     PCL::STANDARD-GENERIC-FUNCTION-P
                     PCL::STANDARD-READER-METHOD-P
                     PCL::STANDARD-METHOD-P
                     PCL::COMPUTE-EFFECTIVE-SLOT-DEFINITION-INITARGS
                     PCL::COMPUTE-DEFAULT-INITARGS
                     PCL::|SETF PCL CLASS-SLOT-VALUE|
                     PCL::METHOD-COMBINATION-TYPE PCL::STANDARD-CLASS-P
                     PCL::LEGAL-SPECIALIZER-P
                     PCL::COMPUTE-SLOT-ACCESSOR-INFO
                     PCL::STANDARD-BOUNDP-METHOD-P
                     PCL::RAW-INSTANCE-ALLOCATOR
                     PCL::|SETF PCL SLOT-DEFINITION-DEFSTRUCT-ACCESSOR-SYMBOL|
                     PCL::|SETF PCL CLASS-INITIALIZE-INFO|
                     PCL::COMPUTE-DISCRIMINATING-FUNCTION-ARGLIST-INFO
                     PCL::STANDARD-WRITER-METHOD-P
                     PCL::CLASS-INCOMPATIBLE-SUPERCLASS-LIST
                     PCL::WRAPPER-FETCHER
                     PCL::METHOD-COMBINATION-DOCUMENTATION
                     PCL::|SETF PCL SLOT-DEFINITION-INITARGS|
                     PCL::REMOVE-BOUNDP-METHOD
                     PCL::|SETF PCL CLASS-DEFSTRUCT-CONSTRUCTOR|
                     PCL::SHORT-COMBINATION-OPERATOR
                     PCL::REMOVE-NAMED-METHOD
                     PCL::LEGAL-DOCUMENTATION-P
                     PCL:CLASS-DIRECT-SUPERCLASSES
                     PCL:CLASS-DIRECT-SUBCLASSES
                     PCL:CLASS-DIRECT-DEFAULT-INITARGS
                     PCL:SLOT-DEFINITION-READERS
                     PCL:SLOT-VALUE-USING-CLASS
                     PCL:COMPUTE-APPLICABLE-METHODS PCL:CLASS-NAME
                     PCL:CLASS-PROTOTYPE PCL:READER-METHOD-CLASS
                     PCL:REMOVE-METHOD PCL:SLOT-DEFINITION-INITFORM
                     PCL:UPDATE-INSTANCE-FOR-REDEFINED-CLASS
                     PCL:UPDATE-INSTANCE-FOR-DIFFERENT-CLASS
                     PCL:CHANGE-CLASS PCL:METHOD-FUNCTION
                     PCL:DIRECT-SLOT-DEFINITION-CLASS
                     PCL:MAKE-METHOD-LAMBDA
                     PCL:EFFECTIVE-SLOT-DEFINITION-CLASS
                     PCL:CLASS-SLOTS PCL:COMPUTE-SLOTS
                     PCL:SLOT-DEFINITION-NAME PCL:FINALIZE-INHERITANCE
                     PCL:GENERIC-FUNCTION-LAMBDA-LIST
                     PCL:CLASS-DIRECT-SLOTS PCL:CLASS-DEFAULT-INITARGS
                     PCL:COMPUTE-DISCRIMINATING-FUNCTION
                     PCL:CLASS-FINALIZED-P PCL:GENERIC-FUNCTION-NAME
                     PCL:REMOVE-DEPENDENT
                     PCL:COMPUTE-CLASS-PRECEDENCE-LIST
                     PCL:ADD-DEPENDENT PCL:SLOT-BOUNDP-USING-CLASS
                     PCL:ACCESSOR-METHOD-SLOT-DEFINITION
                     PCL:SHARED-INITIALIZE PCL:ADD-DIRECT-METHOD
                     PCL:SLOT-DEFINITION-LOCATION
                     PCL:SLOT-DEFINITION-INITFUNCTION
                     PCL:SLOT-DEFINITION-ALLOCATION PCL:ADD-METHOD
                     PCL:GENERIC-FUNCTION-METHOD-CLASS
                     PCL:METHOD-SPECIALIZERS
                     PCL:SLOT-DEFINITION-INITARGS
                     PCL:WRITER-METHOD-CLASS PCL:ADD-DIRECT-SUBCLASS
                     PCL:SPECIALIZER-DIRECT-METHODS
                     PCL:GENERIC-FUNCTION-METHOD-COMBINATION
                     PCL:ALLOCATE-INSTANCE PCL:COMPUTE-EFFECTIVE-METHOD
                     PCL:SLOT-DEFINITION-TYPE PCL:SLOT-UNBOUND
                     PCL:INITIALIZE-INSTANCE PCL:FUNCTION-KEYWORDS
                     PCL:REINITIALIZE-INSTANCE PCL:VALIDATE-SUPERCLASS
                     PCL:GENERIC-FUNCTION-METHODS
                     PCL:REMOVE-DIRECT-METHOD PCL:METHOD-LAMBDA-LIST
                     PCL:MAKE-INSTANCE
                     PCL:COMPUTE-EFFECTIVE-SLOT-DEFINITION
                     PCL:PRINT-OBJECT PCL:METHOD-QUALIFIERS
                     PCL:METHOD-GENERIC-FUNCTION
                     PCL:REMOVE-DIRECT-SUBCLASS
                     PCL:MAKE-INSTANCES-OBSOLETE
                     PCL:SLOT-MAKUNBOUND-USING-CLASS
                     PCL:ENSURE-GENERIC-FUNCTION-USING-CLASS
                     PCL:SLOT-MISSING PCL:MAP-DEPENDENTS
                     PCL:FIND-METHOD-COMBINATION
                     PCL:ENSURE-CLASS-USING-CLASS
                     PCL:NO-APPLICABLE-METHOD
                     PCL:SLOT-DEFINITION-WRITERS
                     PCL:COMPUTE-APPLICABLE-METHODS-USING-CLASSES
                     PCL:CLASS-PRECEDENCE-LIST))
  (SETF (GET PCL::V 'COMPILER::PROCLAIMED-CLOSURE) T)) 
