
(in-package "DSYS")

(defsystem inspecter
    ()
  (:module clos pcl (:type :system))
  (:module clx clx (:type :system))
  (:module clcs clcs (:type :system))
  (:parallel
   "package"
   (:load "precom")
   "item"
   "pane"
   "inspect"
   "input"
   "object" 
   "other"
   (:compile "precom")
   (:forms :compile 
	    (progn #+cmu (purify))
	    :load
	    (progn #+cmu (purify)))))

(defparameter *inspecter-files*
  '((("systems") "lisp"
     "inspecter")
    (("rose3" "inspecter") "lisp"
     "sysdef"
     "package" "item" "pane" "inspect" "input" "object" "other" "precom")
    (("rose3" "inspecter") "txt"
     "rgb")))
