
(in-package "USER")

(unless (and (find-package "DSYS")
	     (find-symbol "INITIALIZE" "DSYS")
	     (fboundp (find-symbol "INITIALIZE" "DSYS")))
  (proclaim '(special *initialize-systems-p*))
  (setq *initialize-systems-p* nil)
  (load "/local1/lisp-build/systems/system")
  (setq *initialize-systems-p* t))

(dsys:initialize-systems :system-only-p t)

(dsys:compile-system 'pcl)
