;;; -*- Mode: Lisp; Package: DSYS; Base: 10; Syntax: Common-Lisp -*-

(in-package "DSYS")

(set-system-source-file 'common-lisp (subfile '("common-lisp") :name "sysdef"))

(pushnew 'common-lisp *auto-load-systems*)
