(in-package :user)
(load #-kcl "system" 
      #+kcl (if #-|NeXT| (probe-file "system.o") #+|NeXT| nil
		"system.o" "system.lisp"))
(dolist (sn '(system pcl clcs common-lisp clx inspecter))
  (push (cons (dsys::find-system sn) :compile) dsys::*loaded-systems*))
;(format t "~&Type:~%(si:save-system \"systems_kcl\")~%")
;#+kcl(si:save-system "systems_kcl")
;#+lucid(lucid:disksave "sys" :full-gc t)
