;;; -*- Mode: Lisp; Package: DSYS; Base: 10; Syntax: Common-Lisp -*-

(in-package "DSYS")

(set-system-source-file 'clcs (subfile '("clcs") :name "sysdef"))

#+akcl (pushnew 'clcs *auto-load-systems*)
