
(in-package 'inspect)

(eval-when (compile) (class-of *package*))

(proclaim #+testing dsys::*testing-declaration* 
	  #-testing dsys::*fast-declaration*)

(pcl::precompile-random-code-segments inspect)

