;;; -*- Mode: Lisp; Base: 10; Syntax: Common-Lisp; Package:(COMMON-LISP (LISP DSYS CONDITIONS PCL)) -*-
;;; File: common-lisp/package.lisp 
;;; Author: Richard Harris

(lisp:in-package "COMMON-LISP" :use '("LISP"))

(shadowing-import dsys:*dsys-shadowing-import-symbols*)

(use-package '("PCL" "CONDITIONS" "DSYS"))

#+cmu
(when (find-package "PRETTY-PRINT")
  (shadow '(GENERIC-FLET GENERIC-LABELS WITH-ADDED-METHODS)
	  "PRETTY-PRINT"))

#+cmu
(when (find-package "XLIB")
  (shadow '(DEFCLASS DEFMETHOD MAKE-INSTANCE PRINT-OBJECT WITH-SLOTS)
	  "XLIB"))

(when (find-package "XP")
  (funcall (intern "INSTALL" "XP")  :macro nil :shadow t))

(when (find-package "SERIES")
  (funcall (intern "INSTALL" "SERIES") :macro t :shadow t :implicit-map t))
